package com.alterdekim.game.message.amf;

import com.alterdekim.game.message.SerializerUtils;
import com.alterdekim.game.message.serializer.BinaryMessageSerializer;
import com.alterdekim.game.utils.ByteUtils;
import com.alterdekim.game.utils.StringUtils;
import lombok.AllArgsConstructor;
import lombok.Getter;
import lombok.NoArgsConstructor;
import lombok.Setter;
import lombok.extern.slf4j.Slf4j;

import java.lang.reflect.Field;
import java.nio.ByteBuffer;
import java.util.*;

@Slf4j
public class AMFDeserializer {
    public static List<AMFObject> deserialize(List<Byte> bytes) {
        List<AMFObject> objects = new ArrayList<>();
        DeserializerSession session = new DeserializerSession();
        for( int i = 0; i < bytes.size(); i++ ) {
            switch (session.getState()) {
                case VALUE_TYPE:
                    session.getCurrent().setType(AMFValueType.toValueType(bytes.get(i)));
                    session.setState(DeserializerState.VALUE_DATA);
                    i = parseLength(bytes, i, session);
                    break;
                case VALUE_DATA:
                    if(session.getCurrent().getType() == AMFValueType.NULL) {
                        session.getCurrent().setValue(null);
                        session.setState(DeserializerState.VALUE_TYPE);
                        session.getData().clear();
                        objects.add(session.getCurrent());
                        session.setCurrent(new AMFObject());
                    }
                    session.getData().add(bytes.get(i));
                    if(session.getData().size() >= session.getValueLength()) {
                        session.getCurrent().setValue(deserializeSingleValue(session.getCurrent().getType(), session.getData()));
                        session.getData().clear();
                        session.setState(DeserializerState.VALUE_TYPE);
                        objects.add(session.getCurrent());
                        session.setCurrent(new AMFObject());
                    }
                    break;
            }
        }
        return objects;
    }

    private static DeserializeSignleResult deserializeSingle( List<Byte> bytes ) {
        DeserializerSession session = new DeserializerSession();
        for( int i = 0; i < bytes.size(); i++ ) {
            switch (session.getState()) {
                case VALUE_TYPE:
                    session.getCurrent().setType(AMFValueType.toValueType(bytes.get(i)));
                    session.setState(DeserializerState.VALUE_DATA);
                    i = parseLength(bytes, i, session);
                    break;
                case VALUE_DATA:
                    if(session.getCurrent().getType() == AMFValueType.NULL) {
                        session.getCurrent().setValue(null);
                        session.setState(DeserializerState.VALUE_TYPE);
                        session.getData().clear();
                        return new DeserializeSignleResult(session.getCurrent(), i);
                    }
                    session.getData().add(bytes.get(i));
                    if(session.getData().size() >= session.getValueLength()) {
                        session.getCurrent().setValue(deserializeSingleValue(session.getCurrent().getType(), session.getData()));
                        session.getData().clear();
                        session.setState(DeserializerState.VALUE_TYPE);
                        return new DeserializeSignleResult(session.getCurrent(), i);
                    }
                    break;
            }
        }
        return new DeserializeSignleResult(new AMFObject(), 0);
    }

    private static int bytes2int(List<Byte> bytes) {
        int value = 0;
        for (byte b : bytes) {
            value = (value << 8) + (b & 0xFF);
        }
        return value;
    }

    private static int parseArrayLength(List<Byte> bytes, int i) {
        for( int u = i; u < bytes.size(); u++ ) {
            if( bytes.get(u) == 0x00 &&
                    bytes.get(u+1) == 0x00 &&
                    bytes.get(u+2) == 0x09 ) {
                return u+2;
            }
        }
        return 0;
    }

    private static int parseObjectLength(List<Byte> bytes, int i) {
        List<Integer> pos = new ArrayList<>();
        int ignore = 0;
        for( int u = i; u < bytes.size()-2; u++ ) {
            if( bytes.get(u) == 0x00 &&
                    bytes.get(u+1) == 0x00 &&
                    bytes.get(u+2) == 0x09 ) {
                if( ignore <= 0 ) {
                    pos.add(u + 2);
                    u += 2;
                    continue;
                }
                ignore--;
            } else if( u != i &&
                    bytes.size() > u+3 &&
                    bytes.get(u) == 0x03 &&
                    bytes.get(u+3) >= 0x30 &&
                    bytes.get(u+3) <= 0x7A ) {
                if( BinaryMessageSerializer.deserializeLongUnsafe(bytes.subList(u+1, u+3)).intValue() <= (bytes.size() - u) ) {
                    ignore++;
                }
            }
        }
        return pos.get(0);
    }

    private static int parseLength(List<Byte> bytes, int i, DeserializerSession session) {
        switch (session.getCurrent().getType()) {
            case NUMBER:
                session.setValueLength(8);
                break;
            case BOOLEAN:
                session.setValueLength(1);
                break;
            case STRING:
                List<Byte> len = bytes.subList(i+1, i+3);
                session.setValueLength(BinaryMessageSerializer.deserializeLongUnsafe(len).intValue());
                return i + 2;
            case OBJECT:
                session.setValueLength(parseObjectLength(bytes, i) - i);
                break;
            case NULL:
                session.getCurrent().setValue(null);
                session.getData().clear();
                session.setState(DeserializerState.VALUE_TYPE);
                break;
            case MIXED_ARRAY:
                session.setValueLength(parseArrayLength(bytes, i) - i);
                break;
        }
        return i;
    }

    private static Object deserializeSingleValue(AMFValueType type, List<Byte> bytes) {
        return switch (type) {
            case NUMBER -> Double.longBitsToDouble(BinaryMessageSerializer.deserializeLong(bytes));
            case BOOLEAN -> bytes.get(0) == 0x01;
            case STRING -> new String(SerializerUtils.bytesToPrimitive(bytes));
            case OBJECT -> deserializeObject(bytes);
            case MIXED_ARRAY -> deserializeArray(bytes);
            default -> null;
        };
    }

    private static List<AMFObject> deserializeArray(List<Byte> bytes) {
        List<AMFObject> arr = new ArrayList<>();
        int cnt = 1;
        for( int i = 4; i < bytes.size(); i++ ) {
            if(bytes.get(i) == 0x00 && bytes.get(i+1) == 0x00 && bytes.get(i+2) == 0x09) cnt--;
            if( cnt <= 0 ) break;
            i += 2;
            DeserializeSignleResult r = deserializeSingle(bytes.subList(i, bytes.size()));
            arr.add(r.getObject());
            i += r.getI();
        }
        return arr;
    }

    private static Map<String, AMFObject> deserializeObject(List<Byte> bytes) {
        Map<String, AMFObject> map = new HashMap<>();
        ObjectDeserializerState state = ObjectDeserializerState.KEY;
        String key = "";
        for( int i = 0; i < bytes.size(); i++ ) {
            if(bytes.get(i) == 0x00 && bytes.get(i+1) == 0x00 && bytes.get(i+2) == 0x09) break;
            switch (state) {
                case KEY:
                    if(i+2 > bytes.size()) return map;
                    List<Byte> key_len_bytes = bytes.subList(i, i+2);
                    int keyLen = BinaryMessageSerializer.deserializeLongUnsafe(key_len_bytes).intValue();
                    if( i+2+keyLen >= bytes.size() ) return map;
                    key = new String(SerializerUtils.bytesToPrimitive(bytes.subList(i+2, i+2+keyLen)));
                    state = ObjectDeserializerState.VALUE;
                    i += 1+keyLen;
                    break;
                case VALUE:
                    DeserializeSignleResult r = deserializeSingle(bytes.subList(i, bytes.size()));
                    map.put(key, r.getObject());
                    i += r.getI();
                    state = ObjectDeserializerState.KEY;
                    break;
            }
        }
        return map;
    }

    public static <T> T deserialize(List<AMFObject> objects, Class<T> valueType) throws Exception {
        List<Field> fields = SerializerUtils.getPrivateFields(valueType);
        T obj = valueType.getDeclaredConstructor().newInstance();
        for( Field field : fields ) {
            AMFObject object = objects.stream()
                    .filter(o -> checkTypeEquality(o.getType(), field.getType().getCanonicalName()))
                    .findFirst()
                    .orElse(null);
            if( object == null ) continue;
            field.set(obj, object.getValue());
            objects.remove(object);
        }
        return obj;
    }

    private static boolean checkTypeEquality(AMFValueType type, String typeName) {
        return switch(type) {
            case NUMBER -> typeName.equals("java.lang.Double");
            case BOOLEAN -> typeName.equals("java.lang.Boolean");
            case STRING -> typeName.equals("java.lang.String");
            case OBJECT -> typeName.equals("java.util.Map");
            case MIXED_ARRAY -> typeName.equals("java.util.List");
            default -> false;
        };
    }

    private enum ObjectDeserializerState {
        KEY,
        VALUE
    }

    @Getter
    @Setter
    private static class DeserializerSession {
        private DeserializerState state;
        private AMFObject current;
        private List<Byte> data;
        private int valueLength;

        private DeserializerSession() {
            state = DeserializerState.VALUE_TYPE;
            current = new AMFObject();
            data = new ArrayList<>();
            valueLength = 0;
        }
    }

    private enum DeserializerState {
        VALUE_TYPE,
        VALUE_DATA
    }

    @AllArgsConstructor
    @NoArgsConstructor
    @Getter
    @Setter
    private static class DeserializeSignleResult {
        private AMFObject object;
        private int i;
    }
}
